""" TRANSACT 

@author: Soufiane Mourragui <soufiane.mourragui@gmail.com>

Main method for the domain adaptation step.

Example
-------

    
Notes
-------


References
-------


"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy
import pandas as pd
from sklearn.metrics.pairwise import kernel_metrics
from joblib import Parallel, delayed

from sklearn.model_selection import GridSearchCV, KFold
from sklearn.linear_model import Ridge, ElasticNet, Lasso, LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.base import clone

from transact.pv_computation import PVComputation
from transact.interpolation import Interpolation
from transact.kernel_computer import KernelComputer


class TRANSACT:
    """
    Perform the whole domain adaptation step, i.e.:
        - Perform KernelPCA on source and target independently.
        - Align kernel principal components geometrically to find directions
        that are shared across source and target. This yields Principal 
        Vectors (PV).
        - Interpolate between source and target principal vectors to balance
        effect of source and target.
        - Compute the optimal interpolation time, i.e. time that minimizes the
        discrepancies in marginal between source and target.
        - Out-of-sample extension: project new dataset onto the consensus feature.

    Features to be added in future:
        - Prediction. Train classifier on source projected.
        - Vizualisation: tSNE plot of consensus feature.
        - Permutation test: how significant are the canonical angles between 
        principal vectors.
        - Ad-hoc kernel that can be passed as functional to use kernels not 
        already implemented.

    Parameters
    ----------
    kernel : str
        Name of the kernel to be used in the whole algorithm. Has to be compliant with
        scikit-learn (https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.kernel_metrics.html#sklearn.metrics.pairwise.kernel_metrics)
    
    kernel_params : dict
        Parameters of the kernel (degree for polynomial kernel, gamma for RBF). 
        Naming has to be compliant with scikit-learn.

    n_components : int or dict
        Number of components for kernel PCA. If int, then same number of components
        for source and target. If dict, then must be of the form {'source':int,
        'target':int} with indication of number of components for source and target.

    n_pv : int
        Number of principal vectors.

    method : str, default to 'two-stage'
        Method used for computing the principal vectors. Only 'two-stage' has been implemented.

    step: int, default to 100
        Number of interpolation steps.

    n_jobs: int, default to 1
        Number of concurrent threads to use for tasks that can be parallelized.

    verbose: bool, default to False (can be int)
        Degree of verbosity.
    """

    def __init__(self,
                kernel='linear',
                kernel_params=None,
                n_components=None,
                n_pv=None,
                method='two-stage',
                step=100,
                n_jobs=1,
                verbose=False):

        self.kernel = kernel
        self.kernel_params_ = kernel_params or {}
        self.kernel_values_ = KernelComputer(self.kernel, self.kernel_params_)

        self.source_data_ = None
        self.target_data_ = None

        self.is_fitted = False

        self.n_components = n_components
        self.n_pv = n_pv
        self.method = method
        self.step = step

        self.predictive_clf = None

        self.n_jobs = n_jobs
        self.verbose = verbose


    def fit(self,
            source_data,
            target_data,
            n_components=None,
            n_pv=None,
            method='two-stage',
            step=100,
            with_interpolation=True,
            left_center=True):

        """
        Fit the domain adaptive classifier, i.e.:
            - Compute Gram matrix (kernel_values_).
            - Compute principal vectors (principal_vectors_).
            - Interpolate between the PVs (interpolation_).
            - Find optimal interpolation time.

        Parameters
        ----------
        source_data : np.ndarray of shape (n_source_samples, n_features), dtype=float
            source data, matrix with samples in the rows. pandas.DataFrame would
            be supported.

        target_data : np.ndarray of shape (n_target_samples, n_features), dtype=float
            target data, matrix with samples in the rows. pandas.DataFrame would
            be supported.
            WARNING: features need to be ordered in the same way as in source_data.

        n_components: int, default to None
            Number of components. If not set in __init__, and default values, then
            maximum number of principal components for source and target will be considered.

        n_pv: int, default to None
            Number of Principal Vectors. If not set in __init__, and default values, then
            maximum number of PV will be computed.

        method : str, default to 'two-stage'
            Method used for computing the principal vectors. Only 'two-stage' has been implemented.

        step: int, default to 100
            Number of interpolation steps.

        with_interpolation: bool, default to True
            Bool indicating whether interpolation shall also be fitted. Useful for just computing PV
            prior to null distribution fitting (and choose of PV number).

        left_center: bool, default to True
            Bool indicating whether the output should be mean-centered, i.e. whether source and target
            consensus features values (or PVs if no interpolation) must have an independent mean-centering.

        Returns
        -------
        self : TRANSACT
            Fitted instance.
        """

        # Save parameters
        self.source_data_ = source_data
        self.target_data_ = target_data
        self.method = method or self.method
        self.n_components = n_components or self.n_components
        self.n_pv = n_pv or self.n_pv
        self.step = step or self.step
        self.left_center = left_center

        # Compute kernel values
        self.kernel_values_.fit(source_data, target_data, center=False)

        # Compute principal vectors
        self.principal_vectors_ = PVComputation(self.kernel, self.kernel_params_)
        self.principal_vectors_.fit(self.source_data_,
                                    self.target_data_,
                                    method=self.method,
                                    n_components=self.n_components,
                                    n_pv=self.n_pv)

        # Stop here if interpolation should not be computed.
        if not with_interpolation:
            return self

        # Set up interpolation scheme
        self.interpolation_ = Interpolation(self.kernel, self.kernel_params_)
        self.interpolation_.fit(self.principal_vectors_, self.kernel_values_)

        # Compute optimal interpolation time
        self._compute_optimal_time(step=self.step, left_center=self.left_center)

        self.is_fitted = True

        return self


    def null_distribution_pv_similarity(self, method='gene_shuffling', n_iter=100):
        """
        Generate a null distribution for the PV similarity function.
            - Gene shuffling: genes get shuffled in source to destroy any structure existing
            at the gene-level while preserving the sample structure. PV get recomputed and 
            similarity is saved.

        Parameters
        ----------
        method : string, default to gene_shuffling
            Method used for generating the null distribution.
            Only method developped: gene_shuffling

        n_iter: int, default to 100
            Number of iterations

        Returns
        -------
        np.ndarray, dtype=float, shape (n_iter, n_pv)
            Array containing the distribution of similarity after shuffling. Each row
            contains the values of one shuffling across PVs.
        """

        if method.lower() == 'gene_shuffling':
            null_method = self._gene_shuffling
        else:
            raise NotImplementedError('%s is not a proper method for generating null distribution'%(method))

        null_distribution = Parallel(n_jobs=self.n_jobs, verbose=self.verbose)\
                                    (delayed(null_method)() for _ in range(n_iter))

        return np.array(null_distribution)


    def _gene_shuffling(self):
        perm = np.random.permutation(self.source_data_.shape[1])
        pv = PVComputation(self.kernel, self.kernel_params_)
        pv.fit(self.source_data_[:,perm],
            self.target_data_,
            method=self.method,
            n_components=self.n_components,
            n_pv=self.n_pv)

        return np.cos(pv.canonical_angles)


    def fit_predictor(self, X, y, alpha_values=None, l1_ratio=0.5):
        """
        Project X on consensus features and train a predictor for y

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features), dtype=float
            Dataset to project. Features should be ordered in same way as in source_data
            and target_data.

        y : np.ndarray of shape (n_samples, 1), dtype=float
            Output to predict

        Returns
        -------
        """
        self.alpha_values = alpha_values if alpha_values is not None else np.logspace(-10,5,34)
        self.l1_ratio_values = [0.2, 0.4, 0.5, 0.6, 0.8, 0.9, 1.]
        param_grid ={
            'regression__alpha': self.alpha_values,
            'regression__l1_ratio': self.l1_ratio_values
        }

        #Grid search setup
        self.predictive_clf = GridSearchCV(Pipeline([
                                ('scaler', StandardScaler(with_mean=True, with_std=False)),
                                ('regression', ElasticNet())
                                ]),\
                                cv=10,
                                n_jobs=self.n_jobs,
                                param_grid=param_grid,
                                verbose=self.verbose,
                                scoring='neg_mean_squared_error')
        self.predictive_clf.fit(self.transform(X, center=False), y)

        return self


    def compute_pred_performance(self, X, y, cv=10):
        """
        Compute predictive performance of predictive model by cross-validation
        on X and y.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features), dtype=float
            Dataset to project. Features should be ordered in same way as in source_data
            and target_data.

        Returns
        -------
        np.ndarray of shape (n_samples, n_pv), dtype=float
            Dataset projected on consensus features.
        """

        kf = KFold(n_splits=cv, shuffle=True)
        X_projected = self.transform(X)

        if self.predictive_clf is None:
            print('BEWARE: NOT FITTED INSTANCE')
            self.fit_predictor(X,y)
        clf = clone(self.predictive_clf)

        y_predicted = np.zeros(X.shape[0])
        for train_index, test_index in kf.split(X_projected):
            clf.fit(X_projected[train_index], y[train_index])
            y_predicted[test_index] = clf.predict(X_projected[test_index])

        return scipy.stats.pearsonr(y_predicted, y)


    def predict(self, X):
        return self.predictive_clf.predict(self.transform(X, center=False))


    def transform(self, X, center=False):
        """
        Project a dataset X onto the consensus features.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features), dtype=float
            Dataset to project. Features should be ordered in same way as in source_data
            and target_data.

        Returns
        -------
        np.ndarray of shape (n_samples, n_pv), dtype=float
            Dataset projected on consensus features.
        """
        return self.interpolation_.transform(X, self.optimal_time, center)


    def _compute_optimal_time(self, step=100, left_center=True):
        # Based on Kolmogorov Smirnov statistics, find interpolation time

        # Compute the interpolated values
        interpolated_values = Parallel(n_jobs=self.n_jobs, verbose=self.verbose)\
                            (delayed(self.interpolation_.project_data)(s/step, center=left_center)
                                for s in range(step+1))
        interpolated_values = np.array(interpolated_values).transpose(2,0,1)
        source_interpolated_values = interpolated_values[:,:,:self.source_data_.shape[0]]
        target_interpolated_values = interpolated_values[:,:,self.source_data_.shape[0]:]

        self.optimal_time = []
        self.ks_statistics = []
        self.ks_p_values = []

        # For each PV, find the time when interpolation has the largest overlap.
        for source_pv, target_pv in zip(source_interpolated_values, target_interpolated_values):
            self.ks_statistics.append([])
            for s, t in zip(source_pv, target_pv):
                self.ks_statistics[-1].append(scipy.stats.ks_2samp(s,t))
            self.ks_statistics[-1] = list(zip(*self.ks_statistics[-1]))
            self.ks_p_values.append(self.ks_statistics[-1][-1])
            self.ks_statistics[-1] = self.ks_statistics[-1][0]
            self.optimal_time.append(np.argmin(self.ks_statistics[-1])/step)

        # Save the different statistics
        self.optimal_time = np.array(self.optimal_time) # Optimal tau for each PV.
        self.ks_statistics = np.array(self.ks_statistics) # Computed KS statistics between each PV.
        self.ks_p_values = np.array(self.ks_p_values) # Corresponding p_values.


import numpy as np
import pandas as pd
import pytest
from joblib import Parallel, delayed
from sklearn.datasets import make_spd_matrix
from sklearn.decomposition import PCA

from transact.interpolation import Interpolation
from transact.matrix_operations import _left_center_kernel
from transact.pv_computation import PVComputation
from transact.kernel_computer import KernelComputer


n_source_samples = 150
n_target_samples = 100
n_dimensions = 200
n_iter_spd_cov = 50
cov_sim_ratio = 0.9
n_jobs = 10
n_pc = 20
n_pv = 15
n_tau_comparison = 10
rbf_params = {
    'gamma':1/n_dimensions
}

def generate_covariance_matrix():
    spd_matrices = Parallel(n_jobs=n_jobs, verbose=1)(delayed(make_spd_matrix)(n_dimensions) for _ in range(n_iter_spd_cov))
    return np.mean(spd_matrices, axis=0)


@pytest.fixture(scope='module')
def common_cov_matrix():
    return generate_covariance_matrix()


@pytest.fixture(scope='module')
def source_cov_matrix(common_cov_matrix):
    return cov_sim_ratio * common_cov_matrix + (1-cov_sim_ratio)*generate_covariance_matrix()


@pytest.fixture(scope='module')
def target_cov_matrix(common_cov_matrix):
    return cov_sim_ratio * common_cov_matrix + (1-cov_sim_ratio)*generate_covariance_matrix()


@pytest.fixture(scope='module')
def source_data(source_cov_matrix):
    X = np.random.multivariate_normal(mean=np.zeros(n_dimensions),
                                     cov=source_cov_matrix,
                                     size=n_source_samples)
    return X


@pytest.fixture(scope='module')
def source_data_centered(source_data):
    return source_data - np.mean(source_data, 0)


@pytest.fixture(scope='module')
def target_data(target_cov_matrix):
    X = np.random.multivariate_normal(mean=np.zeros(n_dimensions),
                                     cov=target_cov_matrix,
                                     size=n_target_samples)
    return X


@pytest.fixture(scope='module')
def target_data_centered(target_data):
    return target_data - np.mean(target_data, 0)


@pytest.fixture(scope='module')
def source_PCA(source_data):
    return PCA(n_pc).fit(source_data)


@pytest.fixture(scope='module')
def target_PCA(target_data):
    return PCA(n_pc).fit(target_data)


@pytest.fixture(scope='module')
def source_PC_projected(source_data):
    return PCA(n_pc).fit_transform(source_data)


@pytest.fixture(scope='module')
def target_PC_projected(target_data):
    return PCA(n_pc).fit_transform(target_data)


@pytest.fixture(scope='module')
def linear_PV(source_PCA, target_PCA):
    M = source_PCA.components_.dot(target_PCA.components_.T)
    u,s,v = np.linalg.svd(M)
    return u.T.dot(source_PCA.components_), v.dot(target_PCA.components_)


@pytest.fixture(scope='module')
def source_linear_PV_projected(linear_PV, source_data):
    return source_data.dot(linear_PV[0].T)[:,:n_pv], source_data.dot(linear_PV[1].T)[:,:n_pv]


@pytest.fixture(scope='module')
def target_linear_PV_projected(linear_PV, target_data):
    return target_data.dot(linear_PV[0].T)[:,:n_pv], target_data.dot(linear_PV[1].T)[:,:n_pv]

@pytest.fixture(scope='module')
def rbf_principal_vectors(source_data, target_data):
    return PVComputation('rbf', rbf_params, n_components=n_pc, n_pv=n_pv).fit(source_data, target_data)

@pytest.fixture(scope='module')
def target_rbf_PV_projected(rbf_principal_vectors, target_data):
    return rbf_principal_vectors.transform(target_data)

@pytest.fixture(scope='module')
def source_rbf_PV_projected(rbf_principal_vectors, source_data):
    return rbf_principal_vectors.transform(source_data)

@pytest.fixture(scope='module')
def rbf_kernel_matrix(source_data, target_data):
    k = KernelComputer('rbf', rbf_params)
    k.fit(source_data, target_data, center=False)

    return k

@pytest.fixture(scope='module')
def linear_principal_vectors(source_data, target_data):
    return PVComputation('linear', n_components=n_pc, n_pv=n_pv).fit(source_data, target_data)

@pytest.fixture(scope='module')
def linear_kernel_matrix(source_data, target_data):
    k = KernelComputer('linear')
    k.fit(source_data, target_data, center=False)

    return k

class TestPVComputation():

    @pytest.fixture(scope='class')
    def interpolation(self):
        return Interpolation('linear')

    @pytest.fixture(scope='class')
    def interpolation_rbf(self, rbf_principal_vectors, rbf_kernel_matrix):
        clf_interpol = Interpolation('rbf', kernel_params=rbf_params)
        return clf_interpol.fit(rbf_principal_vectors, rbf_kernel_matrix, n_pv=n_pv)


    @pytest.fixture(scope='class')
    def interpolation_linear(self,
                                    linear_principal_vectors,
                                    linear_kernel_matrix):
        clf = Interpolation('linear')
        return clf.fit(linear_principal_vectors, linear_kernel_matrix, n_pv=n_pv)


    def test_linear_zero_tau(self,
                            interpolation_linear,
                            source_linear_PV_projected,
                            target_linear_PV_projected):

        X = interpolation_linear.project_data(0, source_only=True, target_only=False, center=False)
        np.testing.assert_almost_equal(X, source_linear_PV_projected[0])

        X = interpolation_linear.project_data(0, source_only=False, target_only=True, center=False)
        np.testing.assert_almost_equal(X, target_linear_PV_projected[0])

    def test_linear_one_tau(self,
                            interpolation_linear,
                            source_linear_PV_projected,
                            target_linear_PV_projected):

        X = interpolation_linear.project_data(1, source_only=True, target_only=False, center=False)
        np.testing.assert_almost_equal(X, source_linear_PV_projected[1])

        X = interpolation_linear.project_data(1, source_only=False, target_only=True, center=False)
        np.testing.assert_almost_equal(X, target_linear_PV_projected[1])


    def test_rbf_zero_tau(self,
                        interpolation_rbf,
                        source_rbf_PV_projected,
                        target_rbf_PV_projected):

        X = interpolation_rbf.project_data(0, source_only=True, target_only=False, center=False)
        np.testing.assert_almost_equal(X, source_rbf_PV_projected['source'])

        X = interpolation_rbf.project_data(0, source_only=False, target_only=True, center=False)
        np.testing.assert_almost_equal(X, target_rbf_PV_projected['source'])


    def test_rbf_one_tau(self,
                        interpolation_rbf,
                        source_rbf_PV_projected,
                        target_rbf_PV_projected):

        X = interpolation_rbf.project_data(1, source_only=True, target_only=False, center=False)
        np.testing.assert_almost_equal(X, source_rbf_PV_projected['target'])

        X = interpolation_rbf.project_data(1, source_only=False, target_only=True, center=False)
        np.testing.assert_almost_equal(X, target_rbf_PV_projected['target'])
        

    def test_transform_linear(self,
                        interpolation_linear,
                        source_data,
                        target_data):
        # Test that transform and project_data gives same result for source and target data
        # Here specific to linear scenario.
        for _ in range(n_tau_comparison):
            tau = np.random.uniform(size=n_pv)
            # Without centering
            np.testing.assert_almost_equal(interpolation_linear.project_data(tau, source_only=True, target_only=False, center=False),
                                        interpolation_linear.transform(source_data, tau=tau, center=False))
            np.testing.assert_almost_equal(interpolation_linear.project_data(tau, source_only=False, target_only=True, center=False),
                                        interpolation_linear.transform(target_data, tau=tau, center=False))

            # With centering
            np.testing.assert_almost_equal(interpolation_linear.project_data(tau, source_only=True, target_only=False, center=True),
                                        interpolation_linear.transform(source_data, tau=tau, center=True))
            np.testing.assert_almost_equal(interpolation_linear.project_data(tau, source_only=False, target_only=True, center=True),
                                        interpolation_linear.transform(target_data, tau=tau, center=True))


    def test_transform_rbf(self,
                        interpolation_rbf,
                        source_data,
                        target_data):
        # Test that transform and project_data gives same result for source and target data
        # Here specific to linear scenario.
        for _ in range(n_tau_comparison):
            tau = np.random.uniform(size=n_pv)
            # Without centering
            np.testing.assert_almost_equal(interpolation_rbf.project_data(tau, source_only=True, target_only=False, center=False),
                                        interpolation_rbf.transform(source_data, tau=tau, center=False))
            np.testing.assert_almost_equal(interpolation_rbf.project_data(tau, source_only=False, target_only=True, center=False),
                                        interpolation_rbf.transform(target_data, tau=tau, center=False))

            # With centering
            np.testing.assert_almost_equal(interpolation_rbf.project_data(tau, source_only=True, target_only=False, center=True),
                                        interpolation_rbf.transform(source_data, tau=tau, center=True))
            np.testing.assert_almost_equal(interpolation_rbf.project_data(tau, source_only=False, target_only=True, center=True),
                                        interpolation_rbf.transform(target_data, tau=tau, center=True))



""" PVComputation

@author: Soufiane Mourragui

Example
-------
    
Notes
-------

References
-------

"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import kernel_metrics
from sklearn.decomposition import KernelPCA
from sklearn.metrics.pairwise import kernel_metrics
from sklearn.model_selection import GridSearchCV
from sklearn.linear_model import Ridge, ElasticNet

from transact.matrix_operations import _sqrt_matrix, _center_kernel, _right_center_kernel, _left_center_kernel
from transact.kernel_computer import KernelComputer


class PVComputation:
    """

    Attributes
    -------
    """

    def __init__(self, kernel, kernel_params={}, n_components=None, n_pv=None):

        self.gamma_coef = None
        self.alpha_coef = None
        self.beta_coef = None
        self.canonical_angles = None

        self.kernel = kernel
        self.kernel_ = kernel_metrics()[kernel]
        self.kernel_params_ = kernel_params

        self.kernel_values_ = KernelComputer(self.kernel, self.kernel_params_)

        # Put n_components in dictionary format.
        self.n_components = n_components
        if type(self.n_components) == int:
            self.n_components = {
                s:self.n_components for s in ['source', 'target']
            }
        self.n_pv = n_pv

        self.n_jobs = 1

    def fit(self, source_data, target_data, method='two-stage', n_components=None, n_pv=None):
        """
        Computes the kernel principal vectors between source and target data.

        Parameters
        -------
        source_data: numpy.ndarray, shape (n_samples, n_genes)
            Source data

        target_data: numpy.ndarray, shape (n_samples, n_genes)
            Source data

        method: str, default to "two-stage"
            Method used for computed the kernel PVs, either "two-stage" (first kernel PCA, then
            alignment), or "direct" (direct minimization)

        n_components: int, default to None
            Number of components taken into the decomposition.

        Returned Values
        -------
        self: returns an instance of self.
        """

        # Compute kernel matrices
        self.kernel_values_.fit(source_data, target_data, center=True)

        if method == 'two-stage':
            self._two_stage_computation(n_components, n_pv)
        elif method == 'direct':
            self._direct_computation(n_components)

        return self
        

    def transform(self, X, right_center=False):
        """
        Project data X on source and target kernel principal vectors

        Parameters
        -------
        X: numpy.ndarray, shape (n_samples, n_genes)
            Data to project

        right_center: Boolean, default to False
            Whether data should be implicitly mean centered

        Returned Values
        -------
        Dictionary with 'source' and 'target' as keys, and projected arrays as values.
        """

        X_projected = {}
        for t in ['source', 'target']:
            X_projected[t] = self._project_PV_from_data(X, t, right_center)

        return X_projected


    def fit_transform(self, source_data, target_data, method='two-stage', n_components=None, n_pv=None):
        """
        Computes the kernel principal vectors between source and target data.

        -------
        source_data: numpy.ndarray, shape (n_samples, n_genes)
            Source data

        target_data: numpy.ndarray, shape (n_samples, n_genes)
            Source data

        method: str, default to "two-stage"
            Method used for computed the kernel PVs, either "two-stage" (first kernel PCA, then
            alignment), or "direct" (direct minimization)

        n_components: int or dictionary, default to None
            Number of components taken into account for PCA. Can be int (if same number of components
            for source or target) or dictionary with {'source': int, 'target':int} indicating the
            number of source and target principal components.

        Returned Values
        -------
        source_projected: dictionary

        target_projected: dictionary
        """

        self.fit(source_data, target_data, method, n_components)

        source_projected = {
            'source': self._project_PV_from_data(source_data, 'source'),
            'target': self._project_PV_from_data(source_data, 'target')
        }
        
        target_projected = {
            'source': self._project_PV_from_data(target_data, 'source'),
            'target': self._project_PV_from_data(target_data, 'target')
        }

        return source_projected, target_projected


    def _two_stage_computation(self, n_components=None, n_pv=None):

        self.n_components = n_components or self.n_components
        if self.n_components is None or type(self.n_components) == int:
            self.n_components = {
                s:self.n_components for s in ['source', 'target']
            }

        self.n_pv = n_pv or (self.n_pv or min(self.n_components.values()))

        ## First step: Kernel PCA
        self._dim_reduction()
        
        ## Second step: Align based on cosine similarity
        self._align_principal_components()


    def _dim_reduction(self):
        self.dim_reduc_clf_ = {}
        self.alpha_coef = {}

        # Independent processing of source and target
        for t in ['source', 'target']:
            # Reduce dimensionality using kernelPCA.
            self.dim_reduc_clf_[t] = KernelPCA(self.n_components[t],
                                            kernel=self.kernel,
                                            n_jobs=self.n_jobs,
                                            **self.kernel_params_)
            self.dim_reduc_clf_[t].fit(self.kernel_values_.data[t])

            # Save kernel PCA coefficients
            self.alpha_coef[t] = self.dim_reduc_clf_[t].alphas_ / np.sqrt(self.dim_reduc_clf_[t].lambdas_)


    def _align_principal_components(self):
        self.cosine_similarity_ = self.alpha_coef['source'].T.dot(self.kernel_values_.k_st).dot(self.alpha_coef['target'])
        
        beta_s, theta, beta_t = np.linalg.svd(self.cosine_similarity_)
        self.beta_coef = {}
        self.beta_coef['source'] = beta_s
        self.beta_coef['target'] = beta_t.T # Due to definition of SVD by matplotlib

        # Computation of gamma coefficients
        self.gamma_coef = {}
        for t in ['source', 'target']:
            self.gamma_coef[t] = self.beta_coef[t].T.dot(self.alpha_coef[t].T)
            self.gamma_coef[t] = self.gamma_coef[t][:self.n_pv]

        # Canonical angles
        self.canonical_angles = np.arccos(theta[:self.n_pv])


    def _direct_computation(self, n_components=None):
        print('BEWARE')
        print('DIRECT COMPUTATION HAS NOT BEEN CORRECTED')
        print('AND COME WITH SUBTANTIAL COMPUTATION PROBLEM')
        print('DO NOT USE AS SUCH')
        self.n_components = n_components or min(self.kernel_values_.source_data.shape[0], self.kernel_values_.target_data.shape[0])

        # Compute kernel cosine similarity
        self.kernel_cosine_similarity_ = np.linalg.pinv(_sqrt_matrix(self.kernel_values_.k_s)).dot(self.kernel_values_.k_st).dot(np.linalg.pinv(_sqrt_matrix(self.kernel_values_.k_t)))

        sigma_s, theta, sigma_t = np.linalg.svd(self.kernel_cosine_similarity_)
        sigma_t = sigma_t.T
        self.gamma_coef = {}
        self.gamma_coef['source'] = np.linalg.inv(_sqrt_matrix(self.kernel_values_.k_s)).dot(sigma_s[:,:self.n_components]).T
        self.gamma_coef['target'] = np.linalg.inv(_sqrt_matrix(self.kernel_values_.k_t)).dot(sigma_t[:,:self.n_components]).T

        # Canonical angles
        self.canonical_angles = theta

    def _project_PV_from_data(self, X, t, right_center=False):
        """
        Project data X on source and target kernel principal vectors

        -------
        X: numpy.ndarray, shape (n_samples, n_genes)
            Data to project

        t: str
            Type, either 'source' or 'target'

        right_center: Boolean, default to False
            Whether data should be implicitly mean centered

        Returned Values
        -------
        Dictionary with 'source' and 'target' as keys, and projected arrays as values.
        Projected arrays are of size (n_samples, n_pv)
        """
        
        K = self.kernel_(self.kernel_values_.data[t], X, **self.kernel_params_)
        K = _left_center_kernel(K)
        if right_center:
            K = _right_center_kernel(K)

        return self._project_PV_from_kernel(K,t)

    def _project_PV_from_kernel(self, K, t):
        """
        Project kernel X on source and target kernel principal vectors

        -------
        K: numpy.ndarray, shape (n_samples, n_samples)
            Kernel matrix between data from type t and specific dataset.
            Source (or target) samples in the rows (same order as given to the algorithm)
            New dataset samples in the columns 

        t: str
            Type, either 'source' or 'target'

        Returned Values
        -------
        Dictionary with 'source' and 'target' as keys, and projected arrays as values.
        Projected arrays are of size (n_samples, n_pv)
        """

        return self.gamma_coef[t].dot(K).T 


import numpy as np
import pandas as pd
import pytest
from joblib import Parallel, delayed
from sklearn.datasets import make_spd_matrix
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA

from transact.pv_computation import PVComputation


n_source_samples = 30
n_target_samples = 50
n_dimensions = 100
n_iter_spd_cov = 20
cov_sim_ratio = 0.7
n_jobs = 10
n_pc = 15
n_pv = 10
rbf_param = {
    'gamma': 1 / n_dimensions
}

def generate_covariance_matrix():
    spd_matrices = Parallel(n_jobs=n_jobs, verbose=1)(delayed(make_spd_matrix)(n_dimensions) for _ in range(n_iter_spd_cov))
    return np.mean(spd_matrices, axis=0)


@pytest.fixture(scope='module')
def common_cov_matrix():
    return generate_covariance_matrix()


@pytest.fixture(scope='module')
def source_cov_matrix(common_cov_matrix):
    return cov_sim_ratio * common_cov_matrix + (1-cov_sim_ratio)*generate_covariance_matrix()


@pytest.fixture(scope='module')
def target_cov_matrix(common_cov_matrix):
    return cov_sim_ratio * common_cov_matrix + (1-cov_sim_ratio)*generate_covariance_matrix()


@pytest.fixture(scope='module')
def source_data(source_cov_matrix):
    X = np.random.multivariate_normal(mean=np.ones(n_dimensions),
                                     cov=source_cov_matrix,
                                     size=n_source_samples)
    return X - np.mean(X, axis=0)


@pytest.fixture(scope='module')
def target_data(target_cov_matrix):
    X = np.random.multivariate_normal(mean=np.zeros(n_dimensions),
                                     cov=target_cov_matrix,
                                     size=n_target_samples)
    return X - np.mean(X, axis=0)


@pytest.fixture(scope='module')
def source_offset():
    return np.random.randint(-100, 100, size=n_dimensions)

@pytest.fixture(scope='module')
def target_offset():
    return np.random.randint(-100, 100, size=n_dimensions)


@pytest.fixture(scope='module')
def source_offset_data(source_cov_matrix, source_offset):
    X = np.random.multivariate_normal(mean=source_offset,
                                     cov=source_cov_matrix,
                                     size=n_source_samples)
    return X


@pytest.fixture(scope='module')
def target_offset_data(target_cov_matrix, target_offset):
    X = np.random.multivariate_normal(mean=target_offset,
                                     cov=target_cov_matrix,
                                     size=n_target_samples)
    return X


@pytest.fixture(scope='module')
def source_PCA(source_data):
    return PCA(n_pc).fit(source_data)


@pytest.fixture(scope='module')
def target_PCA(target_data):
    return PCA(n_pc).fit(target_data)


@pytest.fixture(scope='module')
def source_PC_projected(source_data):
    return PCA(n_pc).fit_transform(source_data)


@pytest.fixture(scope='module')
def target_PC_projected(target_data):
    return PCA(n_pc).fit_transform(target_data)


@pytest.fixture(scope='module')
def linear_PV(source_PCA, target_PCA):
    M = source_PCA.components_.dot(target_PCA.components_.T)
    u,s,v = np.linalg.svd(M)
    return u.T.dot(source_PCA.components_), v.dot(target_PCA.components_)


@pytest.fixture(scope='module')
def linear_canonical_angle(source_PCA, target_PCA):
    M = source_PCA.components_.dot(target_PCA.components_.T)
    u,s,v = np.linalg.svd(M)
    return s[:n_pv]

@pytest.fixture(scope='module')
def source_linear_PV_projected(linear_PV, source_data):
    return source_data.dot(linear_PV[0].T)[:,:n_pv], source_data.dot(linear_PV[1].T)[:,:n_pv]


@pytest.fixture(scope='module')
def target_linear_PV_projected(linear_PV, target_data):
    return target_data.dot(linear_PV[0].T)[:,:n_pv], target_data.dot(linear_PV[1].T)[:,:n_pv]



class TestPVComputation():


    @pytest.fixture(scope='class')
    def kernel_pv(self,
                source_data,
                target_data):
        clf = PVComputation('linear', n_components=n_pc, n_pv=n_pv)
        return clf.fit(source_data, target_data, 'two-stage')


    @pytest.fixture(scope='class')
    def kernel_pv_offset(self,
                    source_offset_data,
                    target_offset_data):
        clf_offset = PVComputation('linear', n_components=n_pc, n_pv=n_pv)
        return clf_offset.fit(source_offset_data, target_offset_data, 'two-stage')


    @pytest.fixture(scope='class')
    def kernel_pv_offset_centered(self,
                                source_offset_data,
                                target_offset_data):
        """
        Create same classifier than kernel_pv_offset, but first center the data.
        The idea is to see the influence of kernel centering in the linear setting -- which
        must be completly equivalent.
        """
        clf_offset_centered = PVComputation('linear', n_components=n_pc, n_pv=n_pv)
        x_source = StandardScaler(with_mean=True, with_std=False).fit_transform(source_offset_data)
        x_target = StandardScaler(with_mean=True, with_std=False).fit_transform(target_offset_data)
        return clf_offset_centered.fit(x_source, x_target, 'two-stage')


    @pytest.fixture(scope='class')
    def kernel_pv_source_projected(self,
                                source_data,
                                kernel_pv):
        return kernel_pv.transform(source_data)


    @pytest.fixture(scope='class')
    def kernel_pv_target_projected(self,
                                target_data,
                                kernel_pv):
        return kernel_pv.transform(target_data)

    
    @pytest.fixture(scope='class')
    def kernel_pv_source_offset_projected(self,
                                        source_offset_data,
                                        kernel_pv_offset):
        return kernel_pv_offset.transform(source_offset_data)


    @pytest.fixture(scope='class')
    def kernel_pv_target_offset_projected(self,
                                        target_offset_data,
                                        kernel_pv_offset):
        return kernel_pv_offset.transform(target_offset_data)


    @pytest.fixture(scope='class')
    def kernel_pv_source_offset_centered_projected(self,
                                                source_offset_data,
                                                kernel_pv_offset_centered):
        """
        We project the non-centered data to get a good comparison between the two.
        """
        return kernel_pv_offset_centered.transform(source_offset_data)


    @pytest.fixture(scope='class')
    def kernel_pv_target_offset_centered_projected(self,
                                                target_offset_data,
                                                kernel_pv_offset_centered):
        """
        We project the non-centered data to get a good comparison between the two.
        """
        return kernel_pv_offset_centered.transform(target_offset_data)


    def test_data_dim(self, source_data, target_data):
        assert source_data.shape == (n_source_samples, n_dimensions)
        assert target_data.shape == (n_target_samples, n_dimensions)


    def test_coefficients_shape(self, kernel_pv, kernel_pv_offset, kernel_pv_offset_centered):
        assert kernel_pv.gamma_coef['source'].shape == (n_pv, n_source_samples)
        assert kernel_pv.gamma_coef['target'].shape == (n_pv, n_target_samples)

        assert kernel_pv_offset.gamma_coef['source'].shape == (n_pv, n_source_samples)
        assert kernel_pv_offset.gamma_coef['target'].shape == (n_pv, n_target_samples)

        assert kernel_pv_offset_centered.gamma_coef['source'].shape == (n_pv, n_source_samples)
        assert kernel_pv_offset_centered.gamma_coef['target'].shape == (n_pv, n_target_samples)


    def test_dim_reduction(self,
                        source_data,
                        target_data,
                        kernel_pv,
                        source_PC_projected,
                        target_PC_projected):

        source_data_proj = kernel_pv.dim_reduc_clf_['source'].transform(source_data)
        target_data_proj = kernel_pv.dim_reduc_clf_['target'].transform(target_data)

        np.testing.assert_almost_equal(source_data_proj, source_PC_projected)
        np.testing.assert_almost_equal(target_data_proj, target_PC_projected)


    def test_source_projection(self,
                            source_linear_PV_projected,
                            kernel_pv_source_projected):
        np.testing.assert_almost_equal(source_linear_PV_projected[0],
                                    kernel_pv_source_projected['source'])
        np.testing.assert_almost_equal(source_linear_PV_projected[1],
                                    kernel_pv_source_projected['target'])


    def test_target_projection(self,
                            target_linear_PV_projected,
                            kernel_pv_target_projected):
        np.testing.assert_almost_equal(target_linear_PV_projected[0],
                                    kernel_pv_target_projected['source'])
        np.testing.assert_almost_equal(target_linear_PV_projected[1],
                                    kernel_pv_target_projected['target'])


    def test_canonical_angle_value(self,
                                kernel_pv,
                                linear_canonical_angle):
        angles = kernel_pv.canonical_angles
        np.testing.assert_equal(np.sort(angles), angles)
        np.testing.assert_almost_equal(angles, np.arccos(linear_canonical_angle))


    def test_linear_dim_reduction_offset_parameters(self,
                        kernel_pv_offset,
                        kernel_pv_offset_centered):

        np.testing.assert_almost_equal(kernel_pv_offset.dim_reduc_clf_['source'].alphas_,
                                    kernel_pv_offset_centered.dim_reduc_clf_['source'].alphas_)
        np.testing.assert_almost_equal(kernel_pv_offset.dim_reduc_clf_['target'].alphas_,
                                    kernel_pv_offset_centered.dim_reduc_clf_['target'].alphas_)

        np.testing.assert_almost_equal(kernel_pv_offset.dim_reduc_clf_['source'].lambdas_,
                                    kernel_pv_offset_centered.dim_reduc_clf_['source'].lambdas_)
        np.testing.assert_almost_equal(kernel_pv_offset.dim_reduc_clf_['target'].lambdas_,
                                    kernel_pv_offset_centered.dim_reduc_clf_['target'].lambdas_)


    def test_linear_pv_offset_parameters(self,
                            kernel_pv_offset,
                            kernel_pv_offset_centered):

        np.testing.assert_almost_equal(kernel_pv_offset.beta_coef['source'],
                                    kernel_pv_offset_centered.beta_coef['source'])
        np.testing.assert_almost_equal(kernel_pv_offset.beta_coef['target'],
                                    kernel_pv_offset_centered.beta_coef['target'])

        np.testing.assert_almost_equal(kernel_pv_offset.gamma_coef['source'],
                                    kernel_pv_offset_centered.gamma_coef['source'])
        np.testing.assert_almost_equal(kernel_pv_offset.gamma_coef['target'],
                                    kernel_pv_offset_centered.gamma_coef['target'])


    def test_linear_kernel_centering(self, kernel_pv_offset, kernel_pv_offset_centered):
        """
        Test the influence of kernel centering using linear kernel. Centering before of after
        is theoretically equivalent
        """
        np.testing.assert_almost_equal(kernel_pv_offset.kernel_values_.k_s, kernel_pv_offset_centered.kernel_values_.k_s)
        np.testing.assert_almost_equal(kernel_pv_offset.kernel_values_.k_t, kernel_pv_offset_centered.kernel_values_.k_t)
        np.testing.assert_almost_equal(kernel_pv_offset.kernel_values_.k_st, kernel_pv_offset_centered.kernel_values_.k_st)


    def test_linear_source_projection_with_offset(self,
                                            kernel_pv_source_offset_projected,
                                            kernel_pv_source_offset_centered_projected):

        np.testing.assert_almost_equal(kernel_pv_source_offset_centered_projected['source'],
                                    kernel_pv_source_offset_projected['source'])
        np.testing.assert_almost_equal(kernel_pv_source_offset_centered_projected['target'],
                                    kernel_pv_source_offset_projected['target'])


    def test_linear_target_projection_with_offset(self,
                                            kernel_pv_target_offset_projected,
                                            kernel_pv_target_offset_centered_projected):

        np.testing.assert_almost_equal(kernel_pv_target_offset_centered_projected['source'],
                                    kernel_pv_target_offset_projected['source'])
        np.testing.assert_almost_equal(kernel_pv_target_offset_centered_projected['target'],
                                    kernel_pv_target_offset_projected['target'])

    


""" Interpolation

@author: Soufiane Mourragui

Example
-------
    
Notes
-------

References
-------

"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import kernel_metrics
from sklearn.decomposition import KernelPCA
from sklearn.metrics.pairwise import pairwise_kernels, kernel_metrics

from transact.matrix_operations import _sqrt_matrix, _center_kernel, centering_matrix
from transact.pv_computation import PVComputation
from transact.kernel_computer import KernelComputer



class Interpolation:

    def __init__(self, kernel, kernel_params={}):

        self.kernel = kernel
        self.kernel_ = kernel_metrics()[kernel]
        self.kernel_params_ = kernel_params

        self._principal_vectors = None
        self.kernel_values_ = KernelComputer(self.kernel, self.kernel_params_)

    def fit(self,
            principal_vectors,
            kernel_values,
            n_pv=None):

        # Feed values to the interpolation scheme
        self._principal_vectors = principal_vectors
        self.kernel_values_ = kernel_values

        self.n_pv = n_pv or (self._principal_vectors.n_pv or min(self._principal_vectors.n_components.values()))

        self._compute_coef_matrix()

        return self

    def project_data(self, tau, source_only=False, target_only=False, center=True):
        """
        Compute the source and/or target data projected on the interpolated path at time
        tau.

        Parameters
        -------

        Returned Values
        -------

        """

        if type(tau) == int or type(tau) == float:
            tau = [tau]*self.n_pv
        tau = np.array(tau)

        sample_coef_matrix = self.pv_matrix_
        if center:
            sample_coef_matrix = self.centering_matrix_.dot(sample_coef_matrix)

        if source_only:
            return sample_coef_matrix[:self.n_source_samples].dot(self._angular_interpolation_matrix(tau))
        elif target_only:
            return sample_coef_matrix[self.n_source_samples:].dot(self._angular_interpolation_matrix(tau))
        else:
            return sample_coef_matrix.dot(self._angular_interpolation_matrix(tau))


    def transform(self, X, tau, center=True):
        
        if type(tau) == int or type(tau) == float:
            tau = [tau]*self.n_pv
        tau = np.array(tau)
        
        K = self.kernel_values_.transform(X, center=center, right_center=True)
        return K.dot(self.coef_matrix_).dot(self._angular_interpolation_matrix(tau))


    def _angular_interpolation_matrix(self, tau):
        assert tau.shape == (self.n_pv,)

        return np.block([[np.diag(self._gamma_interpolations(tau))],
                        [np.diag(self._xi_interpolations(tau))]])


    def _gamma_interpolations(self, tau):
        num = np.sin((1-tau)*self._principal_vectors.canonical_angles)
        den = np.sin(self._principal_vectors.canonical_angles)
        return num / den


    def _xi_interpolations(self, tau):
        num = np.sin(tau*(self._principal_vectors.canonical_angles))
        den = np.sin(self._principal_vectors.canonical_angles)
        return num / den


    def _compute_coef_matrix(self):
        """
        Compute different matrices of used later:
        - kernel_matrix_: kernel values between source and target data (stacked).
        - centering_matrix_: global centering matrix.
        - coef_matrix_: coefficients for source and target principal components
        - pv_matrix_: source and target principal vectors coefficients later used for interpolation.
        """
        gamma_source = self._principal_vectors.gamma_coef['source']
        gamma_target = self._principal_vectors.gamma_coef['target']
        self.n_source_samples = gamma_source.shape[1]
        self.n_target_samples = gamma_target.shape[1]

        self.kernel_matrix_ = np.block([
            [
                self.kernel_values_.k_s,
                self.kernel_values_.k_st
            ],
            [
                self.kernel_values_.k_ts,
                self.kernel_values_.k_t
            ]
        ])

        self.centering_matrix_ = np.block([
            [
                centering_matrix(self.n_source_samples),
                np.zeros((self.n_source_samples, self.n_target_samples))
            ],
            [
                np.zeros((self.n_target_samples, self.n_source_samples)),
                centering_matrix(self.n_target_samples)
            ]
        ])

        self.coef_matrix_ =  np.block([
            [
                gamma_source.T[:,:self.n_pv],
                np.zeros((self.n_source_samples, self.n_pv))
            ],
            [
                np.zeros((self.n_target_samples, self.n_pv)),
                gamma_target.T[:,:self.n_pv]
            ]
        ])

        self.pv_matrix_ = self.kernel_matrix_.dot(self.centering_matrix_).dot(self.coef_matrix_)

    @property
    def principal_vectors(self):
        return self._principal_vectors

    @principal_vectors.setter
    def principal_vectors(self, pv):
        self._principal_vectors = pv

        if not self.is_pv_fitted_:
            self._compute_coef_matrix
        self.is_pv_fitted_ = True



""" Interpolation

@author: Soufiane Mourragui

Example
-------
    
Notes
-------

References
-------

"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import kernel_metrics
from sklearn.decomposition import KernelPCA
from sklearn.metrics.pairwise import pairwise_kernels, kernel_metrics

from transact.matrix_operations import _sqrt_matrix, _center_kernel, _right_center_kernel


class KernelComputer():

    def __init__(self, kernel, kernel_params={}):

        self.kernel = kernel
        self.kernel_ = kernel_metrics()[kernel]
        self.kernel_params_ = kernel_params
        self._source_data = None
        self._target_data = None
        self._data = {}
        self.center=False

        self._empty_kernel_values()


    def fit(self, X_source, X_target, center=True):

        self._source_data = X_source
        self._target_data = X_target
        self._data = {'source': X_source, 'target': X_target}
        self.center = center

        self._compute_kernel(center=self.center)

        return self


    def transform(self, X, center=False, right_center=False):
        """
        Returns kernel matrix with X in rows and (source, target) in columns
        """
        K_with_source = self.kernel_(X, self._source_data, **self.kernel_params_)
        K_with_target = self.kernel_(X, self._target_data, **self.kernel_params_)

        if center:
            K_with_source = _center_kernel(K_with_source)
            K_with_target = _center_kernel(K_with_target)
        elif right_center:
            K_with_source = _right_center_kernel(K_with_source)
            K_with_target = _right_center_kernel(K_with_target)
            
        return np.concatenate([K_with_source, K_with_target], axis=1)


    def _compute_kernel(self, center=False):
        # Global kernel matrix
        self.kernel_matrix_ = self.kernel_(np.concatenate([self._source_data, self._target_data]),
                                            **self.kernel_params_)
        
        # Individual kernel matrices for source, target, and cross-over.
        n_source_samples = self._source_data.shape[0]
        self.k_s = self.kernel_matrix_[:n_source_samples,:n_source_samples]
        self.k_t = self.kernel_matrix_[n_source_samples:,n_source_samples:]
        self.k_st = self.kernel_matrix_[:n_source_samples,n_source_samples:]

        if center:
            self.k_s = _center_kernel(self.k_s)
            self.k_t = _center_kernel(self.k_t)
            self.k_st = _center_kernel(self.k_st)
            
        self.k_ts = self.k_st.T

        self.kernel_submatrices = {
            'source': self.k_s,
            'target': self.k_t,
            'source-target': self.k_st,
            'target-source': self.k_ts
        }


    def _empty_kernel_values(self):

        self.kernel_matrix_ = None
        self.kernel_submatrices = None
        self.k_s = None
        self.k_t = None
        self.k_st = None
        self.k_ts = None


    @property
    def source_data(self):
        return self._source_data

    @source_data.setter
    def source_data(self, X):
        self._source_data = X
        self._data['source'] = X

        if X is None:
            self._empty_kernel_values()
        elif self._target_data is not None:
            self._compute_kernel(center=self.center)

    @property
    def target_data(self):
        return self._target_data

    @target_data.setter
    def target_data(self, X):
        self._target_data = X
        self._data['target'] = X

        if X is None:
            self._empty_kernel_values()
        elif self._source_data is not None:
            self._compute_kernel(center=self.center)

    @property
    def data(self):
        return self._data
    


""" Matrix Operations

@author: Soufiane Mourragui

Example
-------
    
Notes
-------
BEWARE : _center_kernel is set to n_s instead of n_s - 1. Might be problematic.

References
-------

"""

import numpy as np

def _sqrt_matrix(m):
        s,u = np.linalg.eigh(m)
        # Clip at 0 to avoid negative value
        return u.dot(np.diag(np.sqrt(s.clip(0)))).dot(u.T)


def _center_kernel(k):
    n_s = k.shape[0]
    n_t = k.shape[1]
    return centering_matrix(n_s).dot(k).dot(centering_matrix(n_t))


def _right_center_kernel(k):
    n_t = k.shape[1]
    return k.dot(centering_matrix(n_t))


def _left_center_kernel(k):
    n_s = k.shape[0]
    return centering_matrix(n_s).dot(k)

def centering_matrix(n):
    """
    Computes the centering matrix of size n, i.e. the matrix with (n-1)/n on the diagonal
    and 1/n elsewhere, corresponding to a one-side centering of kernel matrix.

    -------
    n: int
        Size of the matrix to center.

    Returned Values
    -------
    C_n: np.ndarray of size (n,n)
        Centering matrix.
    """

    return np.identity(n) - (np.ones(shape=(n,n)) / n)


import numpy as np
import pandas as pd
import pytest
from joblib import Parallel, delayed
from sklearn.datasets import make_spd_matrix
from sklearn.decomposition import PCA

from transact.TRANSACT import TRANSACT


n_source_samples = 150
n_target_samples = 100
n_dimensions = 200
n_iter_spd_cov = 50
cov_sim_ratio = 0.9
n_jobs = 1
n_step = 10
n_iter_null = 20
n_pc = 20
n_pv = 15
rbf_params = {
    'gamma':1/n_dimensions
}

def generate_covariance_matrix():
    spd_matrices = Parallel(n_jobs=n_jobs, verbose=1)(delayed(make_spd_matrix)(n_dimensions) for _ in range(n_iter_spd_cov))
    return np.mean(spd_matrices, axis=0)


@pytest.fixture(scope='module')
def common_cov_matrix():
    return generate_covariance_matrix()


@pytest.fixture(scope='module')
def source_cov_matrix(common_cov_matrix):
    return cov_sim_ratio * common_cov_matrix + (1-cov_sim_ratio)*generate_covariance_matrix()


@pytest.fixture(scope='module')
def target_cov_matrix(common_cov_matrix):
    return cov_sim_ratio * common_cov_matrix + (1-cov_sim_ratio)*generate_covariance_matrix()


@pytest.fixture(scope='module')
def source_data(source_cov_matrix):
    X = np.random.multivariate_normal(mean=np.zeros(n_dimensions),
                                     cov=source_cov_matrix,
                                     size=n_source_samples)
    return X


@pytest.fixture(scope='module')
def source_data_centered(source_data):
    return source_data - np.mean(source_data, 0)


@pytest.fixture(scope='module')
def target_data(target_cov_matrix):
    X = np.random.multivariate_normal(mean=np.zeros(n_dimensions),
                                     cov=target_cov_matrix,
                                     size=n_target_samples)
    return X


class TestTRANSACT():

    @pytest.fixture(scope='class')
    def linear_kernel_transact(self, source_data, target_data):

        linear_clf = TRANSACT('linear', n_jobs=n_jobs)

        return linear_clf.fit(source_data,
                            target_data, 
                            n_components=n_pc,
                            n_pv=n_pv,
                            method='two-stage',
                            step=n_step)


    @pytest.fixture(scope='class')
    def linear_output_source(self, source_data, linear_kernel_transact):
        coef = np.random.randint(-10,10,size=linear_kernel_transact.principal_vectors_.n_pv)
        return linear_kernel_transact.transform(source_data).dot(coef)


    @pytest.fixture(scope='class')
    def linear_null_distribution(self, linear_kernel_transact):
        return linear_kernel_transact.null_distribution_pv_similarity(n_iter=n_iter_null)


    @pytest.fixture(scope='class')
    def rbf_kernel_transact(self, source_data, target_data):

        rbf_clf = TRANSACT('rbf', kernel_params=rbf_params,  n_jobs=n_jobs)

        return rbf_clf.fit(source_data,
                        target_data, 
                        n_components=n_pc,
                        n_pv=n_pv,
                        method='two-stage',
                        step=n_step)


    def test_shape_and_values_linear(self,
                                linear_kernel_transact):
        
        assert linear_kernel_transact.optimal_time.shape == (n_pv,)
        for e in linear_kernel_transact.optimal_time:
            assert e >= 0
            assert e <= 1
        assert linear_kernel_transact.ks_statistics.shape == (n_pv, n_step+1)
        assert linear_kernel_transact.ks_p_values.shape == (n_pv, n_step+1)
        assert np.prod(linear_kernel_transact.ks_p_values >= 0)
        assert np.prod(linear_kernel_transact.ks_p_values <= 1)

        # Check that algo select time of maximal interpolation.
        for i in range(n_pv):
            assert np.prod(linear_kernel_transact.ks_statistics[i] >= 
                linear_kernel_transact.ks_statistics[i][int(linear_kernel_transact.optimal_time[i]*n_step)])

    def test_shape_and_values_rbf(self,
                                rbf_kernel_transact):
        
        assert rbf_kernel_transact.optimal_time.shape == (n_pv,)
        for e in rbf_kernel_transact.optimal_time:
            assert e >= 0
            assert e <= 1
        assert rbf_kernel_transact.ks_statistics.shape == (n_pv, n_step+1)
        assert rbf_kernel_transact.ks_p_values.shape == (n_pv, n_step+1)
        assert np.prod(rbf_kernel_transact.ks_p_values >= 0)
        assert np.prod(rbf_kernel_transact.ks_p_values <= 1)

        # Check that algo select time of maximal interpolation.
        for i in range(n_pv):
            assert np.prod(rbf_kernel_transact.ks_statistics[i] >= 
                rbf_kernel_transact.ks_statistics[i][int(rbf_kernel_transact.optimal_time[i]*n_step)])


    def test_shape_and_values_linear_transform(self,
                                            linear_kernel_transact,
                                            source_data,
                                            target_data):

        X_source = linear_kernel_transact.transform(source_data)
        X_target = linear_kernel_transact.transform(target_data)

        assert X_source.shape == (n_source_samples, n_pv)
        assert X_target.shape == (n_target_samples, n_pv)


    def test_shape_and_values_rbf_transform(self,
                                            rbf_kernel_transact,
                                            source_data,
                                            target_data):

        X_source = rbf_kernel_transact.transform(source_data)
        X_target = rbf_kernel_transact.transform(target_data)

        assert X_source.shape == (n_source_samples, n_pv)
        assert X_target.shape == (n_target_samples, n_pv)


    def test_shape_linear_null_distribution(self, linear_null_distribution):
        assert linear_null_distribution.shape == (n_iter_null, n_pv)
        # Test monotonicity
        for i in range(n_pv-1):
            assert np.prod(linear_null_distribution[:,i] >= linear_null_distribution[:,i+1])


    def test_prediction(self, source_data, linear_kernel_transact, linear_output_source):
        linear_kernel_transact.fit_predictor(source_data, linear_output_source)
        np.testing.assert_almost_equal(linear_output_source, linear_kernel_transact.predict(source_data))

    def test_prediction(self, source_data, linear_kernel_transact, linear_output_source):
        linear_kernel_transact.fit_predictor(source_data, linear_output_source)
        pred_perf = linear_kernel_transact.compute_pred_performance(source_data, linear_output_source, cv=3)
        
        assert pred_perf[0] > 0.99



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="transact_dr",
    version="1.0.1",
    author="Soufiane Mourragui <soufiane.mourragui@gmail.com>, ",
    author_email="soufiane.mourragui@gmail.com",
    description="TRANSACT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NKI-CCB/TRANSACT",
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'scipy', 'pandas', 'matplotlib', 'scikit-learn'],
    python_requires='>=3.6',
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Development Status :: 1 - Planning",
    ),
)


import numpy as np
import pandas as pd
import pytest
from joblib import Parallel, delayed

from transact.kernel_computer import KernelComputer
from transact.matrix_operations import _left_center_kernel, _center_kernel


n_source_samples = 30
n_target_samples = 50
n_dimensions = 400

@pytest.fixture(scope='module')
def source_data():
    return np.random.normal(0,1,size=(n_source_samples, n_dimensions))


@pytest.fixture(scope='module')
def target_data():
    return np.random.normal(0,1,size=(n_target_samples, n_dimensions))



class TestKernelComputer():

    @pytest.fixture(scope='class')
    def linear_kernel_computer(self):
        return KernelComputer('linear')

    @pytest.fixture(scope='class')
    def uncentered_linear_kernel_computer(self, source_data, target_data):
        return KernelComputer('linear').fit(source_data, target_data, center=False)

    @pytest.fixture(scope='class')
    def rbf_kernel_computer(self):
        return KernelComputer('rbf')


    @pytest.fixture(scope='class')
    def uncentered_rbf_kernel_computer(self, source_data, target_data):
        return KernelComputer('rbf').fit(source_data, target_data, center=False)


    def test_source_data_property_no_target(self,
                                        source_data,
                                        linear_kernel_computer):

        linear_kernel_computer.source_data = source_data

        assert linear_kernel_computer.kernel_matrix_ is None
        assert linear_kernel_computer.kernel_submatrices is None
        assert linear_kernel_computer.k_s is None
        assert linear_kernel_computer.k_t is None
        assert linear_kernel_computer.k_st is None
        assert linear_kernel_computer.k_ts is None

        linear_kernel_computer.source_data = None


    def test_source_data_property_with_target(self,
                                        source_data,
                                        target_data,
                                        linear_kernel_computer):

        linear_kernel_computer.target_data = target_data
        linear_kernel_computer.source_data = source_data

        assert linear_kernel_computer.kernel_matrix_ is not None
        assert linear_kernel_computer.kernel_submatrices is not None
        assert linear_kernel_computer.k_s is not None
        assert linear_kernel_computer.k_t is not None
        assert linear_kernel_computer.k_st is not None
        assert linear_kernel_computer.k_ts is not None

        linear_kernel_computer.target_data = None
        linear_kernel_computer.source_data = None


    def test_target_data_property_no_source(self,
                                        target_data,
                                        linear_kernel_computer):

        linear_kernel_computer.target_data = target_data

        assert linear_kernel_computer.kernel_matrix_ is None
        assert linear_kernel_computer.kernel_submatrices is None
        assert linear_kernel_computer.k_s is None
        assert linear_kernel_computer.k_t is None
        assert linear_kernel_computer.k_st is None
        assert linear_kernel_computer.k_ts is None

        linear_kernel_computer.target_data = None


    def test_target_data_property_with_source(self,
                                        source_data,
                                        target_data,
                                        linear_kernel_computer):

        linear_kernel_computer.source_data = source_data
        print(linear_kernel_computer._source_data)
        linear_kernel_computer.target_data = target_data

        assert linear_kernel_computer.kernel_matrix_ is not None
        assert linear_kernel_computer.kernel_submatrices is not None
        assert linear_kernel_computer.k_s is not None
        assert linear_kernel_computer.k_t is not None
        assert linear_kernel_computer.k_st is not None
        assert linear_kernel_computer.k_ts is not None

        linear_kernel_computer.target_data = None
        linear_kernel_computer.source_data = None


    def test_linear_fit(self,
                    source_data,
                    target_data,
                    linear_kernel_computer):

        linear_kernel_computer.fit(source_data, target_data)

        assert linear_kernel_computer.kernel_matrix_.shape == (n_source_samples+n_target_samples, n_source_samples+n_target_samples)
        assert len(linear_kernel_computer.kernel_submatrices.keys()) == 4
        assert 'source' in linear_kernel_computer.kernel_submatrices.keys()
        assert 'target' in linear_kernel_computer.kernel_submatrices.keys()
        assert 'source-target' in linear_kernel_computer.kernel_submatrices.keys()
        assert 'target-source' in linear_kernel_computer.kernel_submatrices.keys()
        assert linear_kernel_computer.k_s.shape == (n_source_samples, n_source_samples)
        assert linear_kernel_computer.k_t.shape == (n_target_samples, n_target_samples)
        assert linear_kernel_computer.k_st.shape == (n_source_samples, n_target_samples)
        assert linear_kernel_computer.k_ts.shape == (n_target_samples, n_source_samples)

    def test_rbf_fit(self,
                    source_data,
                    target_data,
                    rbf_kernel_computer):

        rbf_kernel_computer.fit(source_data, target_data)

        assert rbf_kernel_computer.kernel_matrix_.shape == (n_source_samples+n_target_samples, n_source_samples+n_target_samples)
        assert len(rbf_kernel_computer.kernel_submatrices.keys()) == 4
        assert 'source' in rbf_kernel_computer.kernel_submatrices.keys()
        assert 'target' in rbf_kernel_computer.kernel_submatrices.keys()
        assert 'source-target' in rbf_kernel_computer.kernel_submatrices.keys()
        assert 'target-source' in rbf_kernel_computer.kernel_submatrices.keys()
        assert rbf_kernel_computer.k_s.shape == (n_source_samples, n_source_samples)
        assert rbf_kernel_computer.k_t.shape == (n_target_samples, n_target_samples)
        assert rbf_kernel_computer.k_st.shape == (n_source_samples, n_target_samples)
        assert rbf_kernel_computer.k_ts.shape == (n_target_samples, n_source_samples)


    def test_center_linear(self,
                        source_data, 
                        target_data, 
                        linear_kernel_computer, 
                        uncentered_linear_kernel_computer):

        linear_kernel_computer.fit(source_data, target_data, center=True)

        assert np.any(np.not_equal(linear_kernel_computer.k_s,
                                uncentered_linear_kernel_computer.k_s))
        assert np.any(np.not_equal(linear_kernel_computer.k_t,
                                uncentered_linear_kernel_computer.k_t))
        assert np.any(np.not_equal(linear_kernel_computer.k_st,
                                uncentered_linear_kernel_computer.k_st))

        np.testing.assert_almost_equal(_center_kernel(uncentered_linear_kernel_computer.k_s),
                                    linear_kernel_computer.k_s)
        np.testing.assert_almost_equal(_center_kernel(uncentered_linear_kernel_computer.k_t),
                                    linear_kernel_computer.k_t)
        np.testing.assert_almost_equal(_center_kernel(uncentered_linear_kernel_computer.k_st),
                                    linear_kernel_computer.k_st)


    def test_center_rbf(self,
                        source_data, 
                        target_data, 
                        rbf_kernel_computer, 
                        uncentered_rbf_kernel_computer):

        rbf_kernel_computer.fit(source_data, target_data, center=True)

        assert np.any(np.not_equal(rbf_kernel_computer.k_s,
                                uncentered_rbf_kernel_computer.k_s))
        assert np.any(np.not_equal(rbf_kernel_computer.k_t,
                                uncentered_rbf_kernel_computer.k_t))
        assert np.any(np.not_equal(rbf_kernel_computer.k_st,
                                uncentered_rbf_kernel_computer.k_st))

        np.testing.assert_almost_equal(_center_kernel(uncentered_rbf_kernel_computer.k_s),
                                    rbf_kernel_computer.k_s)
        np.testing.assert_almost_equal(_center_kernel(uncentered_rbf_kernel_computer.k_t),
                                    rbf_kernel_computer.k_t)
        np.testing.assert_almost_equal(_center_kernel(uncentered_rbf_kernel_computer.k_st),
                                    rbf_kernel_computer.k_st)

    def test_transform(self,
                    source_data,
                    target_data,
                    rbf_kernel_computer):

        rbf_kernel_computer.fit(source_data, target_data)

        n_source_samples = source_data.shape[0]
        x_target_proj = rbf_kernel_computer.transform(target_data, center=True)
        np.testing.assert_almost_equal(x_target_proj[:,n_source_samples:], rbf_kernel_computer.k_t)
        np.testing.assert_almost_equal(x_target_proj[:,:n_source_samples], rbf_kernel_computer.k_ts)

        n_source_samples = source_data.shape[0]
        x_source_proj = rbf_kernel_computer.transform(source_data, center=True)
        np.testing.assert_almost_equal(x_source_proj[:,n_source_samples:], rbf_kernel_computer.k_st)
        np.testing.assert_almost_equal(x_source_proj[:,:n_source_samples], rbf_kernel_computer.k_s)



import numpy as np
import pytest
from sklearn.datasets import make_spd_matrix

from transact.matrix_operations import _sqrt_matrix
from transact.matrix_operations import _center_kernel
from transact.matrix_operations import _right_center_kernel
from transact.matrix_operations import _left_center_kernel
from transact.matrix_operations import centering_matrix

size = 20

@pytest.fixture(scope='module')
def random_matrix():
    return make_spd_matrix(n_dim=size)

@pytest.fixture(scope='module')
def C_n():
    return centering_matrix(size) 

class TestPVComputation():

    @pytest.fixture(scope='class')
    def sqrt_matrix(self, random_matrix):
        return _sqrt_matrix(random_matrix)

    def test_sqrt_shape(self, sqrt_matrix, random_matrix):
        assert sqrt_matrix.shape == random_matrix.shape

    def test_sqrt_reconstruction(self, sqrt_matrix, random_matrix):
        np.testing.assert_almost_equal(sqrt_matrix.dot(sqrt_matrix),
                                    random_matrix)


    def test_commutativity_centering(self, random_matrix):
        x = _left_center_kernel(random_matrix)
        x = _right_center_kernel(x)
        np.testing.assert_almost_equal(x, _center_kernel(random_matrix)) 

        x = _right_center_kernel(random_matrix)
        x = _left_center_kernel(x)
        np.testing.assert_almost_equal(x, _center_kernel(random_matrix)) 

    def test_properties_centering_matrix(self, C_n):
        # Idempotent
        np.testing.assert_almost_equal(C_n, C_n.dot(C_n))

        # Spectrum is [1]*(n-1) + [0]
        spectrum = np.ones(size)
        spectrum[0] = 0
        np.testing.assert_almost_equal(np.linalg.eigvalsh(C_n), spectrum)

        # Null space is ones
        np.testing.assert_almost_equal(C_n.dot(np.ones(size)), np.zeros(size))



